__version__ = "0.0.9.4"
__app_name__ = "yabtool"
__human_readable_app_name__ = "Yet Another Backup Tool"
__package_description__ = (
    "Yet Another Backup Tool with controlled flow and archives rolling"
)
